@@ -5,7 +5,7 @@
 #
 .First.lib <- function(lib,pkg) {
   library.dynam("gllm",pkg,lib)
-  cat("This is gllm 0.25\n")
+  cat("This is gllm 0.27\n")
 }
 #
 # EM IPF algorithm of Haber AS207
@@ -20,9 +20,10 @@ emgllm <- function(y,s,X,maxit=1000,tol=0.00001) {
   observed.values<-em$y
   fitted.values<-em$f
   full.table<-em$e
-  return(deviance,
-         observed.values,fitted.values,
-         full.table)
+  list(deviance=deviance,
+       observed.values=observed.values,
+       fitted.values=fitted.values,
+       full.table=full.table)
 }
 as207 <- function(y,s,X,maxit,tol)
   .Fortran("gllm",
@@ -49,7 +50,7 @@ as207 <- function(y,s,X,maxit,tol)
 scatter<- function(y,s) {
   S<-matrix(rep(0,length(y)*length(s)),nrow=length(y))
   for(i in 1:length(s)) if (s[i]<=length(y)) S[s[i],i]<-1
-  return(t(S))
+  t(S)
 }
 #
 # Espeland's scoring algorithm
@@ -253,4 +254,136 @@ anova.gllm <- function(object, ..., test = c("Chisq", "none"))
     c("Likelihood ratio tests of Loglinear Models\n")
   out
 }
+#
+# Test for allelic association following Aston and Wilson.
+# 2 locus version
+#
+ld2 <- function(locus1, locus2=NULL) {
+  all.possible.genos <- function(genotypes,gtp.sep="/") {
+    alleles<- sort(unique(unlist(strsplit(genotypes[!is.na(genotypes)], gtp.sep))))
+    gtp<-cbind(rep(alleles,seq(length(alleles),1,-1)),
+               rev(alleles)[sequence(seq(length(alleles),1,-1))])
+    paste(gtp[,1],gtp[,2],sep=gtp.sep)
+  }
+  genofactor <- function(genotypes, gtp.sep="/") {
+    factor(genotypes, levels=all.possible.genos(genotypes))
+  }
+  ng.to.nall<-function(ngenos) (sqrt(1+8*ngenos)-1)/2
 
+  if (is.null(locus2)) {
+    if (is.table(locus1) && length(dim(locus1))==2) {
+      ld.table<-locus1
+    }else if (is.data.frame(locus1)) {
+      ld.table<-table(genofactor(locus1[,1]), genofactor(locus1[,2]))
+    }
+  }else{
+    ld.table<-table(genofactor(locus1), genofactor(locus2))
+  }
+  siz<-dim(ld.table)
+  if (length(siz)==2 && siz[1]>1 && all(siz>1)) {
+    nall1<-ng.to.nall(siz[1])
+    nall2<-ng.to.nall(siz[2])
+    m0<-ld2.model(nall1,nall2,"~a1+a2")
+    m1<-ld2.model(nall1,nall2,"~a1+a2+d")
+    m2<-ld2.model(nall1,nall2,"~a1+a2+p1")
+    m3<-ld2.model(nall1,nall2,"~a1+a2+p1+p2")
+    m4<-ld2.model(nall1,nall2,"~a1+a2+p1+p2+d")
+    m0<-gllm(c(t(ld.table)),m0$s,m0$X,method="hybrid",em.maxit=1,tol=0.00001) 
+    m1<-gllm(c(t(ld.table)),m1$s,m1$X,method="hybrid",em.maxit=1,tol=0.00001) 
+    m2<-gllm(c(t(ld.table)),m2$s,m2$X,method="hybrid",em.maxit=1,tol=0.00001) 
+    m3<-gllm(c(t(ld.table)),m3$s,m3$X,method="hybrid",em.maxit=1,tol=0.00001) 
+    m4<-gllm(c(t(ld.table)),m4$s,m4$X,method="hybrid",em.maxit=1,tol=0.00001) 
+    res<-list(m0=m0,m1=m1,m2=m2,m3=m3,m4=m4)
+    class(res)<-"ld2"
+    res
+  }else{
+    cat("Must have two polymorphic loci\n")
+  }
+}
+print.ld2 <- function(x,...) {
+  cat("\nAssuming HWE\n")
+  print(summary(x$m1))
+  cat("\n")
+  print(anova(x$m0,x$m1))
+  cat("\nModelling HWD\n")
+  print(summary(x$m4))
+  cat("\n")
+  print(anova(x$m0,x$m2,x$m3,x$m4))
+}
+#
+# Produce the filter and design matrices.
+# Geno is not used directly, but documents the ordering of the 
+# observed table of genotype counts
+#
+ld2.model <- function(nall1, nall2, formula="~a1+a2+p1+p2+d") {
+  mkgeno<-function(nall) {
+    lab<-t(outer(1:nall, 1:nall, paste,sep="/"))
+    lab[lower.tri(lab,diag=TRUE)]
+  }
+  ld.terms<-function(nall1, nall2, formula=formula) {
+    nter<-cumsum(c(1,nall1-1,nall2-1, (nall1-1)*(nall1-1),
+                   (nall2-1)*(nall2-1), (nall1-1)*(nall2-1)))
+    ter<-pmatch(unlist(strsplit(formula,"[~+]")),c("a1","a2","p1","p2","d"))
+    ter<-ter[!is.na(ter)]
+    res<-1
+    for(i in ter) res<-append(res,seq(nter[i]+1, nter[i+1]))
+    res           
+  }
+  ng1<-nall1*(nall1+1)/2
+  ng2<-nall2*(nall2+1)/2
+  gam<-nall1*nall1*nall2*nall2
+  gen<-ng1*ng2
+  mod<-nall1+nall2-1+ 
+       (nall1-1)*(nall1-1)+(nall2-1)*(nall2-1)+ 
+       (nall1-1)*(nall2-1)
+  Geno<-matrix(1:gen, nr=ng1, byrow=TRUE)
+  rownames(Geno)<-mkgeno(nall1)
+  colnames(Geno)<-mkgeno(nall2)
+  S<-array(0, dim=c(nall1,nall1,nall2,nall2))
+  X<-matrix(0, nr=gam, nc=mod)
+  colnames(X)<-c("Int",paste("a1",2:nall1,sep="."),
+                       paste("a2",2:nall2,sep="."),
+                       paste("p1",outer(2:nall1,2:nall1,paste,sep=""),sep="."),
+                       paste("p2",outer(2:nall2,2:nall2,paste,sep=""),sep="."),
+                       paste("d",outer(2:nall1,2:nall2,paste,sep=""),sep="."))
+#
+# Produce all nall1^2 * nall2^2 gametes and map onto genotypes
+#
+  k<-0
+  for(i in 1:nall1) for(i2 in 1:i)
+  for(j in 1:nall2) for(j2 in 1:j) {
+    k<-k+1
+    S[i,i2,j,j2]<-S[i,i2,j2,j]<-S[i2,i,j,j2]<-S[i2,i,j2,j]<-k
+  }
+  S<-matrix(S, nr=nall1*nall1, nc=nall2*nall2)
+  rownames(S)<-outer(1:nall1,1:nall1,paste,sep="/") 
+  colnames(S)<-outer(1:nall2,1:nall2,paste,sep="/") 
+#
+# Then the design matrix
+#
+  k<-0
+  hap<-rep(" ",gam)
+  for(i in 1:nall1) for(i2 in 1:nall1)
+  for(j in 1:nall2) for(j2 in 1:nall2) {
+    k<-k+1
+    a1<-rep(0,nall1)
+    a1[i]<-a1[i]+1
+    a1[i2]<-a1[i2]+1
+    p1<-matrix(0, nr=nall1, nc=nall1)
+    p1[i,i2]<-p1[i,i2]+1
+    a2<-rep(0,nall2)
+    a2[j]<-a2[j]+1
+    a2[j2]<-a2[j2]+1
+    p2<-matrix(0, nr=nall2, nc=nall2)
+    p2[j,j2]<-p2[j,j2]+1
+    d<-matrix(0, nr=nall1, nc=nall2)
+    d[i, j2]<-d[i, j2]+1
+    d[i2, j]<-d[i2, j]+1
+    X[k,]<-c(1,a1[-1],a2[-1],c(p1[-1,-1]), c(p2[-1,-1]), c(d[-1,-1]))
+    hap[k]<-paste(i,j,";",i2,j2,sep="")
+  }
+  rownames(X)<-hap
+  s<-c(t(S))
+  names(s)<-hap
+  list(Geno=Geno, s=s, X=X[,ld.terms(nall1,nall2,formula)])
+}
